package stocks;

import java.rmi.server.UnicastRemoteServer;
import java.rmi.server.StubSecurityManager;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

// This class implements a remote server for the StockQuoteServer interface.
// It keeps a table of stocks, and a table of clients indexed by stock.
// When a client wants to receive stock quotes it calls addWatch.
// When a quote is updated, this object invokes the update method
// in every client that is watching that stock.

// Since this program is just a demo, the list of stocks is
// hard-coded and the stock values are updated every 5 seconds
// by a random amount.

public class StockQuoteServerImpl extends UnicastRemoteServer 
implements StockQuoteServer, Runnable
{
// the table of stocks, indexed by stock name
	protected Hashtable stocks;

// the table of clients watching each stock, indexed by stock name
	protected Hashtable stockClients;

// the known stocks for this server

	protected final static StockQuote initialStocks[] = {
		new StockQuote("FUDJ", 50.0, 0.0),
		new StockQuote("SUNW", 90.0, 1.0),
		new StockQuote("FOO", 10.25, -0.5),
		new StockQuote("BAR", 7.25, 0.5),
		new StockQuote("BAZ", 20.75, -3.75),
		new StockQuote("QUUX", 30.0, 2.5)
	};

// The constructor initializes the tables and fills the stocks
// table with the stocks from initialSocks.

	public StockQuoteServerImpl()
	throws java.rmi.RemoteException
	{

// Create the tables
		stocks = new Hashtable();
		stockClients = new Hashtable();

// Initialize the stocks table from initialStocks
		for (int i=0; i < initialStocks.length; i++) {
			stocks.put(initialStocks[i].stock,
				initialStocks[i]);
		}
	}

// addWatch adds a client to the list of clients watching a stock

	public void addWatch(StockQuoteClient client, String stock)
	throws java.rmi.RemoteException, StockQuoteException
	{

// If we don't know about the stock, throw an exception

		if (stocks.get(stock) == null) {
			throw new UnknownStockException(stock);
		}

// Get the container of clients watching this stock
		Vector clients = (Vector) stockClients.get(stock);

// If no clients are watching, create the container
		if (clients == null) {
			clients = new Vector();
			clients.addElement(client);
			stockClients.put(stock, clients);

// Only add the client if it isn't already there. We don't want to
// double-update clients.

		} else if (!clients.contains(client)) {
			clients.addElement(client);
		}
	}

// removeWatch removes a client from the list of clients watching a stock.

	public void removeWatch(StockQuoteClient client, String stock)
	throws java.rmi.RemoteException, StockQuoteException
	{

// If the stock doesn't exist, throw an exception
		if (stocks.get(stock) == null) {
			throw new UnknownStockException(stock);
		}

// Get the list of clients who are watching this stock
		Vector clients = (Vector) stockClients.get(stock);

// If there are no clients, just return
		if (clients == null) {
			return;
		}

// Remove the client from the list
		clients.removeElement(client);
	}

// removeClient removes a client from each list of watched stocks

	public void removeClient(StockQuoteClient client)
	throws java.rmi.RemoteException
	{
		clearClient(client);
	}

// clearClient actually performs the removal of a client
// from the list of clients for each stock.

	protected void clearClient(StockQuoteClient client)
	{
		Enumeration e = stockClients.elements();

		while (e.hasMoreElements()) {
			Vector v = (Vector) e.nextElement();

			v.removeElement(client);
		}
	}

// getStockList returns an array of all the available stocks

	public String[] getStockList()
	throws java.rmi.RemoteException
	{

// Create the array for the stock names
		String[] stockList = new String[stocks.size()];

		Enumeration e = stocks.keys();
		int i = 0;

// Go through each key in the table and copy the key into the array

		while (e.hasMoreElements()) {
			stockList[i++] = (String) e.nextElement();
		}

		return stockList;
	}

// publishQuote sends a stock quote to every client who is watching

	protected void publishQuote(StockQuote quote)
	{

// Get the list of clients for the stock
		Vector v = (Vector) stockClients.get(quote.stock);

// If there are no clients, we're done
		if (v == null) return;

		Enumeration e = v.elements();

// When we get an exception sending a notification to a client, we
// remove the client. We don't do it until we've sent all the
// notifications however. We store them in badClients until then.

		Vector badClients = null;

		while (e.hasMoreElements()) {

			StockQuoteClient client = (StockQuoteClient)
				e.nextElement();

// send the quote to the client
			try {
				client.quote(quote);

// If we get an error, add the client to the list of bad clients

			} catch (java.rmi.RemoteException oops) {
				if (badClients == null) {
					badClients = new Vector();
				}
				badClients.addElement(client);
			}
		}

// If there were any bad clients, remove them

		if (badClients != null) {
			e = badClients.elements();
			while (e.hasMoreElements()) {
				clearClient(
					(StockQuoteClient) e.nextElement());
			}
		}
	}

// generateNewQuotes creates random changes to each stock value

	protected void generateNewQuotes()
	{
		Enumeration e = stocks.elements();

		while (e.hasMoreElements()) {
			StockQuote quote = (StockQuote)e.nextElement();

// change is the product of two numbers between 0 and 1. Squaring
// the numbers gives a bell-curve centered around 0.5.

			double diff = Math.random() * Math.random();

// subtracting from 0.5 shift the bell curve to be from -0.5 to 0.5
// Multiply by 8, take integer remainder, divide by 8 makes the
// fractional part of the number be as small as eights, but no finer.

			diff = 0.5 - Math.rint(diff * 8.0) / 8.0;
				
			quote.change = diff;

			quote.amount += quote.change;

// Send the quote to the clients

			publishQuote(quote);
		}
	}

	public void run()
	{
		while (true)
		{
			generateNewQuotes();

			try {
				Thread.sleep(5000);
			} catch (Exception insomnia) {
			}
		}
	}

	public static void main(String[] args)
	{

// Always use a security manager when doing RMI

		System.setSecurityManager(new StubSecurityManager());

		try {

// Create the server
			StockQuoteServerImpl server =
				new StockQuoteServerImpl();
			Thread serverThread = new Thread(server);
			serverThread.start();

// Register the object with the naming service as "StockQuotes"
			java.rmi.Naming.rebind("StockQuotes", server);

		} catch (Exception e) {
			System.out.println("Got exception: "+e);
			e.printStackTrace();
		}
	}
}
